use chrono::FixedOffset;
use diaryrs::macros::UnwrapOrFatalAble;
use regex::Regex;
use serde::Deserialize;

use crate::args::Config;

use std::{fs, path};

// Management for existing diary entries
// TODO: Proper error handling for all the times we used `?`

/// Overall stats and information about existing diary entries
///
/// Call `::take` to construct an `Inventory` struct from diary entires on disk
#[derive(Debug)]
pub struct Inventory {
    /// Number of entries in the diary
    pub entries_count: usize,

    /// A vector of [`Entry`] representing each diary entry
    ///
    /// Sorted oldest to newest based on enclosed timestamp
    pub entries: Vec<Entry>,

    /// Total word count of all entries
    ///
    /// Does not include YAML front matter or comments
    pub word_count: u64,

    /// Average word count: simply `word_count / entries` rounded to the nearest integer
    pub avg_word_count: f64,
}

/// Represents the front-matter metadata included within each entry. Does not necessarily have to be associated with any real diary entry.
///
/// "Lower-level" (more representative of what is on disk) than an [`Entry`]. For example, it does not do any validation or processing of the timezone
#[derive(Debug, Deserialize)]
pub struct Metadata {
    timestamp: String,
    timezone: String,
}

impl Metadata {
    pub fn read_from_entry(contents: &str) -> Result<Self, Box<dyn std::error::Error>> {
        // Match YAML section
        let yaml_regex = Regex::new(r"(?U)---\n([\S\s]+)---\n")
            .unwrap_or_fatal("Failed to compile YAML regex. Something is very wrong!");

        let matches = yaml_regex
            .captures(&contents)
            .ok_or("YAML front-matter not found!")?;

        let yaml_str = matches
            .get(1)
            .ok_or("YAML front-matter match not found!")?
            .as_str();

        let meta = serde_yaml::from_str::<Metadata>(yaml_str)?;

        Ok(meta)
    }
}

/// Represents a single diary entry (markdown file) and its associated stats/metadata
///
/// Does not include the actual contents of the entry, only stats
///
/// The offset of this entry can be retrieved from the timestamp. The tzdb-style timezone (e.g. `Australia/Sydney`)
/// is separately specified. Using the offset is probably preferred as the offset should already account for regional daylight saving and other oddities.
#[derive(Debug)]
pub struct Entry {
    pub path: path::PathBuf,

    pub word_count: u64,
    pub timestamp: chrono::DateTime<FixedOffset>,
    pub timezone: chrono_tz::Tz,
}

impl Entry {
    pub fn read(path: path::PathBuf) -> Result<Self, Box<dyn std::error::Error>> {
        let contents = fs::read_to_string(&path)?;

        let meta = Metadata::read_from_entry(&contents)?;
        let timestamp = chrono::DateTime::parse_from_rfc3339(&meta.timestamp)?;
        let timezone: chrono_tz::Tz = meta.timezone[0..meta.timezone.len() - 8].parse()?;

        Ok(Self {
            path,
            word_count: count_words(&contents)?,
            timestamp,
            timezone,
        })
    }
}

impl Inventory {
    /// Take inventory of the diary entries
    ///
    /// Reads the diary entries from disks and produces a useful
    /// `Inventory` struct with stats and information about the entries.
    pub fn take(config: &Config) -> Result<Self, Box<dyn std::error::Error>> {
        // Take directory listing
        let dir = fs::read_dir(&config.base_path)?;

        let md_paths: Vec<_> = dir
            .filter_map(|entry| entry.ok())
            .flat_map(|entry| recurse_paths_md(entry))
            .flatten()
            .collect();

        let mut entries = Vec::new();

        for path in md_paths {
            entries.push(Entry::read(path)?)
        }

        // Sort entry oldest to newest
        entries.sort_by(|a, b| a.timestamp.cmp(&b.timestamp));

        let word_count = entries.iter().fold(0, |acc, e| acc + e.word_count);
        let entries_count = entries.len();

        Ok(Self {
            entries_count,
            entries,
            word_count,
            avg_word_count: (word_count as f64 / entries_count as f64).round(),
        })
    }
}

/// Recursively finds every Markdown (*.md) file in a directory
/// and its subdirectories
///
/// Returns a vector of `PathBuf`s to each Markdown file or a boxed `Error`.
fn recurse_paths_md(dir: fs::DirEntry) -> Result<Vec<path::PathBuf>, Box<dyn std::error::Error>> {
    let mut paths = Vec::new();
    let current_path = dir.path();
    let path_type = dir.file_type()?;

    if path_type.is_dir() {
        // List dir, then run this function on each entry
        // Add anything returned to the paths vector
        let dir = fs::read_dir(&current_path)?;
        paths.extend(
            dir.filter_map(|entry| entry.ok())
                .flat_map(|entry: fs::DirEntry| recurse_paths_md(entry))
                .flatten(),
        );
    }

    if path_type.is_file() {
        // Check if file is an md file
        if let Some(ext) = current_path.extension() {
            if ext == "md" {
                paths.push(current_path);
            }
        }
    }

    Ok(paths)
}

fn count_words(contents: &str) -> Result<u64, Box<dyn std::error::Error>> {
    let re_titles = Regex::new(r"(?m)^#{1,6} .+")
        .unwrap_or_fatal("Failed to compile title regex. Something is very wrong!");
    let re_comments = Regex::new(r"(?s-m)<!---?.+-->")
        .unwrap_or_fatal("Failed to compile comment regex. Something is very wrong!");
    let re_images = Regex::new(r"!\[.*\]\(.+\)")
        .unwrap_or_fatal("Failed to compile image regex. Something is very wrong!");

    // Cut YAML header, comments and images
    let contents = &re_titles.replace_all(
        contents
            .split("---")
            .collect::<Vec<&str>>()
            .pop()
            .ok_or("No content found")?,
        "",
    );
    let contents = &re_comments.replace_all(contents, "");
    let contents = &re_images.replace_all(contents, "");

    Ok(contents.split_whitespace().count() as u64)
}
